#!/usr/bin/env python

"""Tests for `w7x` package."""

import unittest


class Test_w7x(unittest.TestCase):
    """Tests for `w7x` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_version_type(self):
        """Assure that version type is str."""
        import w7x
        self.assertIsInstance(w7x.__version__, str)



#!/user/bin/env python
"""
w7x option starter
"""
import os
import numpy as np
import sys
import unittest
import doctest
import pathlib
import argparse
import matplotlib.pyplot as plt

import tfields
import w7x


def run_doctests():
    """
    Find all doctests and execute them
    """
    this_dir = pathlib.Path(__file__).parent.resolve()
    for f in list(this_dir.glob('**/*.py')):
        doctest.testfile(str(f.resolve()),
                         module_relative=False)  # , verbose=True, optionflags=doctest.ELLIPSIS)


def load_unittests(loader=None, suite=None):
    if loader is None:
        loader = unittest.TestLoader()
    if suite is None:
        suite = unittest.TestSuite()
    parent = pathlib.Path(__file__).parent.parent
    for f in list(parent.glob('*/test*.py')):
        sys.path.insert(0, str(f.parent))
        mod = __import__(f.name[:-3])
        for test in loader.loadTestsFromModule(mod):
            suite.addTests(test)
        sys.path.remove(str(f.parent))
    return suite


def run_unittests(arg):
    run_doctests()
    # unittest.main(defaultTest='load_unittests')


def diffuse(args):
    """
    diffuse a extended vmec run and save the result
    """
    raise NotImplementedError("Paths should be set by user.")
    localDir = tfields.lib.in_out.resolve("~/Data/EXTENDER/{args.vmec_id}/".format(**locals()))
    datPath = os.path.join(localDir, "{args.vmec_id}.dat".format(**locals()))

    cyl = w7x.extender.getHybridFromDat(datPath)
    grid = w7x.flt.Grid(cylinder=cyl)
    config = w7x.flt.MagneticConfig.from_dat_file(datPath, grid=grid)
    run = w7x.flt.Run(config)

    saveDir = os.path.join(args.baseDir, "{args.vmec_id}".format(**locals()))
    tfields.lib.in_out.mkdir(saveDir, isDir=True)
    path = tfields.lib.in_out.resolve(os.path.join(saveDir,
                                      '{args.vmec_id}-fld.nest.npz'.format(**locals())))
    tracerConnectionResults, componentLoads, startPoints3D = \
        run.line_diffusion(startPoints=args.startPoints)
    # npt.saveNested(path, (tracerConnectionResults, componentLoads, startPoints3D))
    """ can be read like
    tracerConnectionResults, componentLoads, startPoints3D = npt.readNested(path)
    """


def poincare(args):
    ''' poincare plot creation '''
    phiList = args.phi.values

    if args.phi.deg:
        phiList = [val / 180 * np.pi for val in phiList]

    if not args.assemblies.off:
        machine = w7x.flt.Machine.from_mm_ids(*args.assemblies.values)
    else:
        machine = w7x.flt.Machine()

    relativeCurrents = args.magneticConfig.relativeCurrents
    datPath = args.magneticConfig.path
    if datPath:
        datPath = tfields.lib.in_out.resolve(datPath)
        cyl = w7x.extender.getHybridFromDat(datPath)
        grid = w7x.flt.Grid(cylinder=cyl)
        magneticConfig = w7x.flt.MagneticConfig.from_dat_file(datPath,
                                                                 grid=grid)
    elif relativeCurrents:
        magneticConfig = w7x.flt.MagneticConfig.createWithCurrents(
            relativeCurrents=relativeCurrents)
    else:
        magneticConfig = w7x.flt.MagneticConfig.createWithCurrents()

    ''' plotting '''
    axis = tfields.plotting.gca(2)
    for phi in phiList:
        axis.grid(color='lightgrey')
        machine.plot_poincare(phi, axis=axis)
        magneticConfig.plot_poincare(phi, axis=axis)
        tfields.plotting.save("~/tmp/poincare-{phi:.4f}".format(**locals()), 'png', 'pgf', 'pdf')
        axis.clear()


def partition_divertor(args):
    """
    cluster the strikeline into the divertor tiles
    """
    from clusterDivertorPartitions import getTileLoadsFromDiffusion, plotTileLoads
    saveDir = os.path.join(args.baseDir, "{args.vmec_id}".format(**locals()))
    path = tfields.lib.in_out.resolve(os.path.join(saveDir,
                                      '{args.vmec_id}-fld.nest.npz'.format(**locals())))
    loads = getTileLoadsFromDiffusion(path)
    plotTileLoads(loads, vmin=0, vmax=60000)
    plt.show()
    


if __name__ == '__main__':
    # create the top-level parser
    parser = argparse.ArgumentParser(prog='W7-X tools app')
    subparsers = parser.add_subparsers(help='sub-command help')

    # create the parser for the "test" command
    parserExtend = subparsers.add_parser('test', help='test help')
    parserExtend.set_defaults(func=run_unittests)

    # create the parser for the "extend" command
    parserExtend = subparsers.add_parser('extend', help='extend help')
    parserExtend.add_argument('vmec_id', type=str, help='vmec_id to extend')
    parserExtend.set_defaults(func=w7x.extender.extend)

    # create the parser for the "diffuse" command
    parserDiffuse = subparsers.add_parser('diffuse', help='diffuse help')
    parserDiffuse.add_argument('vmec_id', type=str,
                               help='already extended vmec_id')
    parserDiffuse.add_argument('--startPoints', type=int,
                               help="hit points = 2 * "
                               "startPoints (forward and backward).",
                               default=12500)
    parserDiffuse.add_argument('--baseDir', type=str,
                               default="~/Data/Strikeline/",
                               help='already extended vmec_id')
    parserDiffuse.set_defaults(func=diffuse)

    # create the parser for the "poincare" command
    parserPoincare = subparsers.add_parser('poincare', help='poincare help')
    parserPoincare.add_argument('--baseDir', type=str,
                                default="~/Data/Strikeline/",
                                help='already extended vmec_id')
    parserPoincare.add_argument('--phi', dest='phi.values', nargs='*', type=float,
                                default=[0.])
    parserPoincare.add_argument('--phi.deg', dest='phi.deg',
                                help='switch phi from radian to degree',
                                action="store_true")
    parserPoincare.add_argument('--assemblies', dest='assemblies.values',
                                nargs='+', type=str,
                                default=w7x.Defaults.Machine.mm_ids)
    parserPoincare.add_argument('--assemblies.off',
                                action="store_true")
    parserPoincare.add_argument('--magneticConfig.relativeCurrents',
                                help='relative currents in case of vacuum config')
    parserPoincare.add_argument('--magneticConfig.coilConfig',
                                help='set the coil config for the relative currents')
    parserPoincare.add_argument('--magneticConfig.path',
                                default=None,
                                help='create config with magnetic field grid at path')
    parserPoincare.set_defaults(func=poincare)

    # create the parser for the "partition_divertor" command
    parserClusterTiles = subparsers.add_parser('partition_divertor',
                                               help='partition_divertor help')
    parserClusterTiles.add_argument('vmec_id', type=str,
                                    help='already extended and diffued vmec_id')
    parserClusterTiles.add_argument('--baseDir', type=str,
                                    default="~/Data/Strikeline/",
                                    help='already extended vmec_id')
    parserClusterTiles.set_defaults(func=partition_divertor)
    
    args = parser.parse_args()
    args.func(args)



import unittest
import w7x
import numpy as np


class MagneticConfig_Test(unittest.TestCase):
    def setUp(self):
        self.inst = w7x.flt.MagneticConfig.default()

    def test_current_units(self):
        units = ['rw', 'Aw', 'A', 'r']
        for unit in units:
            m_tmp = w7x.flt.MagneticConfig.from_currents(
                *self.inst.coil_currents(unit=unit),
                unit=unit, scale=None)
            for check_unit in units:
                self.assertTrue(
                    np.isclose(m_tmp.coil_currents(check_unit),
                               self.inst.coil_currents(check_unit)).all()
                )


class MagneticConfig_FromCurrents_Test(MagneticConfig_Test):
    def setUp(self):
        self.inst = w7x.flt.MagneticConfig.from_currents(
            1, 1, 1, 1, 1,
            -0.23, -0.23,
            0.001, 0.001,
            0.0002, 0.0002, 0.0002, 0.0002, 0.0002,
            unit='rw')


if __name__ == '__main__':
    # m = MagneticConfig_FromCurrents_Test()
    # m.setUp()
    # m.test_current_units()
    unittest.main()



import unittest
import w7x
import os
import shutil
import numpy as np
import pandas as pd

folder = '~/tmp/VMEC/test/'
vmec_features = ["phiedge", "coilCurrents"]
size = 10

magnetic_config = w7x.flt.MagneticConfig.default()
pressure_profile = w7x.vmec.PowerSeries()
current_profile = w7x.vmec.PowerSeries()
magneticAxis = w7x.vmec.SurfaceCoefficients.default_axis()
boundary = w7x.vmec.SurfaceCoefficients.default_boundary()

coil_currents_standard = w7x.Defaults.MagneticConfig.standard_rw
coil_currents_low = [0.6, 0.6, 0.6, 0.6, 0.6, -1.0, -1.0, 0.0, 0.0]
coil_currents_high = [1.4, 1.4, 1.4, 1.4, 1.4, 1.0, 1.0, 0.0, 0.0]


class VmecBatchTest(unittest.TestCase):
    def setUp(self):
        self.batch = w7x.batch.Vmec(folder=folder)
        self.batch.size = size

    def tearDown(self):
        folder = os.path.expanduser(self.batch.folder + self.batch.batch_id)
        shutil.rmtree(os.path.expanduser(folder), ignore_errors=True)

    def test_not_null_batch_id(self):
        self.assertTrue(self.batch.batch_id != None)

    def test_batch_default_size(self):
        self.assertEqual(self.batch.size, size)

    def test_gen_batch_folder(self):
        self.batch.gen()
        batch_folder = os.path.expanduser(folder + self.batch.batch_id)
        self.assertTrue(os.path.isdir(batch_folder))

    def test_gen_runs_folder(self):
        self.batch.gen()
        runs_ids = self.batch.runs
        for run_id in runs_ids:
            run_folder = os.path.expanduser(folder + self.batch.batch_id +
                                            "/" + run_id)
            self.assertTrue(os.path.isdir(run_folder))

    def test_default_magnetic_config(self):
        self.assertEqual(self.batch.magnetic_config.coil_currents(),
                         coil_currents_standard)

    def test_sample_and_set_single_magnetic_config(self):
        self.batch.magnetic_config = (w7x.batch.Distribution(
            np.random.uniform, coil_currents_low, coil_currents_high), 'rw')
        self.assertFalse(self.batch.magnetic_config.coil_currents() ==
                         coil_currents_standard)

    def test_default_pressure_profile(self):
        self.assertEqual(self.batch.pressure_profile.coefficients,
                         w7x.vmec.PowerSeries().coefficients)
        self.assertEqual(self.batch.pressure_profile.ProfileType,
                         w7x.vmec.PowerSeries().ProfileType)

    def test_sample_and_set_single_pressure_profile(self):
        self.batch.pressure_profile = (w7x.batch.Distribution(
            np.random.uniform, [0, 0], [1e-6, -1e-6]), 'power_series')
        self.assertFalse((self.batch.pressure_profile.coefficients ==
                          w7x.vmec.PowerSeries().coefficients).any())

    def test_default_phiedge(self):
        self.assertEqual(self.batch.phiedge,
                         w7x.config.Defaults.VMEC.maxToroidalMagneticFlux)

    def test_sample_and_default_phiedge(self):
        samples = [self.batch.phiedge for i in range(10)]
        for s in samples:
            self.assertEqual(s,
                             w7x.config.Defaults.VMEC.maxToroidalMagneticFlux)

    def test_sample_and_set_single_phiedge(self):
        self.batch.phiedge = w7x.batch.Distribution(np.random.uniform, -3.0,
                                                    -1.0)
        self.assertFalse(self.batch.phiedge ==
                         w7x.config.Defaults.VMEC.maxToroidalMagneticFlux)

    def test_sample_and_set_multiple_phiedge(self):
        self.batch.phiedge = w7x.batch.Distribution(np.random.uniform, -3.0,
                                                    -1.0)
        samples = [self.batch.phiedge for i in range(10)]
        check = [w7x.config.Defaults.VMEC.maxToroidalMagneticFlux
                 ] * len(samples)
        self.assertFalse(
            samples == w7x.config.Defaults.VMEC.maxToroidalMagneticFlux)

    def test_sample_and_set_multiple_relative_phiedge(self):
        self.batch.phiedge = w7x.batch.Distribution(np.random.uniform, -3.0,
                                                    -1.0)
        sample1 = self.batch.phiedge
        sample2 = self.batch.phiedge
        self.assertFalse(sample1 == sample2)

    def test_empty_features_list(self):
        self.batch.gen()
        self.assertRaises(ValueError, self.batch.to_dataframe)

    def test_dataframe_columns(self):
        self.batch.gen()
        self.batch.features = vmec_features
        df = self.batch.to_dataframe()
        self.assertEqual(set(df.columns), set(vmec_features))

    def test_dataframe_len(self):
        self.batch.gen()
        self.batch.features = vmec_features
        df = self.batch.to_dataframe()
        self.assertEqual(len(df), self.batch.size)


if __name__ == '__main__':
    unittest.main()



import rna
import tfields
import w7x
import joblib
import pathlib
import os
import sys
import importlib
import tempfile
import numpy as np
from scipy.stats import poisson
try:
    import cPickle as pickle
except ImportError:
    import pickle


memory = joblib.Memory(location=os.path.join(tempfile.gettempdir(), 'joblib'))


def n_hits_from_load(q, n_tot, area, p_conv, ceta):
    """
    Args:
        q(float) heat load
        n_tot (int): total number of hits
        area (float)
        p_conv (float): convective power
        ceta (int): mapping factor
    """
    p_conv *= 1. / ceta  # correct for mapping all divertors to one
    return int(round((1. * q * n_tot * area) / p_conv))


def load_from_n_hits(n_hits, n_tot, area, p_conv, ceta):
    """
    Args:
        n_hits (int): numer of hits on the area
        n_tot(int): total number of hits
        area (float)
        p_conv (float): convective power
        ceta (int): mapping factor
    """
    p_conv *= 1. / ceta  # correct for mapping all divertors to one
    return (n_hits * p_conv) * 1. / (n_tot * area)


def prob_n_hits_lt_n_des(n_hits_list, n_des_list):
    """
    P(n_hits < n_des | p_conv, n_tot, q_design)
    """
    tile_probs = []
    for n_hits, n_des in zip(n_hits_list, n_des_list):
        n_hits = int(np.round(n_hits))
        p = poisson.cdf(n_des, n_hits)
        tile_probs.append(float(p))
    return tile_probs


@memory.cache
def upper_half(mm_id):
    from sympy.abc import z
    return w7x.flt.MeshedModel.from_mm_id(mm_id).as_Mesh3D().cut(z>0)


class TracerEvaluator(object):
    main_mm_id = None
    mm_ids = None
    def __init__(self):
        self.setup()
        self.clear()

    def setup(self):
        self._mesh = None
        self._design_loads = None
        self._design_loads_local = None
        self._design_loads_average = None
        self._areas = None
        self._areas_local = None
        self._areas_average = None
        self._n_tot = None
        self._p_conv = None
        self._ceta = None

    def clear(self):
        self._conn_len = None
        self._hits = None
        self._n_crits = {}
        self._n_hits = None
        self._n_hits_local = None
        self._n_hits_average = None
        self._measurements = None
        self._measurements_projected = None

    @staticmethod
    def from_loader(loader_file_dir):
        loader_file_dir = pathlib.Path(os.path.abspath(loader_file_dir))
        sys.path.append(str(loader_file_dir.parent))
        module = importlib.import_module('{loader_file_dir.name}.loader'
                                         .format(**locals()))
        evaluator = module.TracerEvaluator()
        return evaluator

    # Environment properties

    @property
    def design_loads(self):
        if self._design_loads is None:
            raise ValueError("Design loads not set")
        return self._design_loads

    @design_loads.setter
    def design_loads(self, values):
        self._design_loads = values

    @property
    def design_loads_local(self):
        if self._design_loads_local is None:
            raise ValueError("Design loads not set")
        return self._design_loads_local

    @design_loads_local.setter
    def design_loads_local(self, values):
        self._design_loads_local = values

    @property
    def design_loads_average(self):
        if self._design_loads_average is None:
            raise ValueError("Design loads not set")
        return self._design_loads_average

    @design_loads_average.setter
    def design_loads_average(self, values):
        self._design_loads_average = values

    @property
    def areas(self):
        if self._areas is None:
            raise ValueError("Areas not set")
        return self._areas

    @areas.setter
    def areas(self, values):
        self._areas = values

    @property
    def areas_local(self):
        if self._areas_local is None:
            raise ValueError("areas_local not set")
        return self._areas_local

    @areas_local.setter
    def areas_local(self, values):
        self._areas_local = values

    @property
    def areas_average(self):
        if self._areas_average is None:
            raise ValueError("areas_average not set")
        return self._areas_average

    @areas_average.setter
    def areas_average(self, values):
        self._areas_average = values

    @property
    def mesh(self):
        """
        tfields.Mesh3D: CAD mesh of the underlying geometry
        """
        if self._mesh is None:
            raise ValueError("Mesh not set")
        return self._mesh

    @mesh.setter
    def mesh(self, value):
        self._mesh = value

    @property
    def ceta(self):
        """
        int :Symmetry mapping factor
        """
        if self._ceta is None:
            raise ValueError("ceta not set")
        return self._ceta

    @ceta.setter
    def ceta(self, value):
        if self._ceta is not None:
            raise ValueError("ceta already set")
        self._ceta = value

    # Experiment properties

    @property
    def measurements(self):
        """
        tfields.TensorFields: measurements at the position of tensors.
            measurements values are stored in fields
        """
        return self._measurements

    @measurements.setter
    def measurements(self, measurements):
        self.clear()
        self._measurements = measurements

    @property
    def measurements_projected(self):
        if self._measurements_projected is None:
            delta = 0.0001
            # speed up projection by filling mesh cache with box_search tree
            self.mesh.tree = self.box_search_tree
            self._measurements_projected = self.mesh.project(self.measurements, delta)
        return self._measurements_projected

    @measurements_projected.setter
    def measurements_projected(self, mp):
        self.clear()
        self._measurements_projected = mp

    # Field Line Diffusion Simulation properties

    @property
    def connection_length(self):
        """
        w7x.flt.ConnectionLenght: Result of a field line tracer diffusion
            process. It contains the hit points (connection_length.hits).
        """
        if self._conn_len is None:
            raise ValueError("connection_length not set")
        return self._conn_len

    @connection_length.setter
    def connection_length(self, conn_len):
        self.clear()
        self._conn_len = conn_len

    @property
    def hits(self):
        """
        hit points in upper module one.
        when setting the hit points, we take care of the mapping to upper module
        one.
        """
        if self._hits is None:
            if self.mm_ids is None:
                raise NotImplementedError("mm_ids not set")
            hits = self.connection_length.hits(*self.mm_ids)
            self.hits = hits
        return self._hits

    @hits.setter
    def hits(self, hits):
        self.clear()
        hits = hits.copy()
        if self.ceta == 10:
            hits.to_segment_one(mirror_z=True)  # map to upper mesh in module 1
        elif self.ceta == 5:
            hits.to_segment_one(mirror_z=False)  # map to mesh in module 1
        elif self.ceta == 1:
            pass
        else:
            raise ValueError("mapping not defined for ceta not in [10, 5, 1]")
        self._hits = hits

    @property
    def n_hits(self):
        """
        Number of hits per section defined by the evaluator.
        This can be calculated automatically if self.hits or
        self.connection_length has been set.
        If the values of n_hits have been calculated before, you can also set
        it. Be aware, that we clear all attributes, if n_hits is set by the
        user.
        """
        if self._n_hits is None:
            self._n_hits = self._eval_n_hits()
        return self._n_hits

    @n_hits.setter
    def n_hits(self, n_hits):
        self.clear()
        self._n_hits = n_hits

    @property
    def n_hits_local(self):
        if self._n_hits_local is None:
            raise ValueError("n_hits_local not set")
        return self._n_hits_local

    @n_hits_local.setter
    def n_hits_local(self, n_hits):
        self.clear()
        self._n_hits_local = n_hits

    @property
    def n_hits_average(self):
        if self._n_hits_average is None:
            raise ValueError("n_hits_average not set")
        return self._n_hits_average

    @n_hits_average.setter
    def n_hits_average(self, n_hits):
        self.clear()
        self._n_hits_local = n_hits

    def _eval_n_hits(self):
        raise NotImplementedError()

    @property
    def n_tot(self):
        if self._n_tot is None:
            raise ValueError()
        return self._n_tot

    @n_tot.setter
    def n_tot(self, value):
        self._n_tot = value
        self._n_crits = {}

    @property
    def p_conv(self):
        """
        float: Convective heat load
        """
        if self._p_conv is None:
            raise ValueError("p_conv is not set")
        return self._p_conv

    @p_conv.setter
    def p_conv(self, value):
        if self._n_tot is None:
            raise ValueError("Please set n_tot first")
        self._n_crits = {}
        self._p_conv = value

    def n_crits(self, variant='local'):
        if variant not in self._n_crits:
            if variant == 'local':
                n_crits = []
                for area, q_crit in zip(self.areas_local,
                                        self.design_loads_local):
                    n_crit = n_hits_from_load(
                        q_crit,
                        self.n_tot,
                        area,
                        p_conv=self.p_conv,
                        ceta=self.ceta)
                    n_crits.append(n_crit)
                self._n_crits[variant] = n_crits
            else:
                raise NotImplementedError()
        return self._n_crits[variant]

    def prob_n_hits_lt_n_des_local(self):
        """
        P(n_hits < n_des | p_conv, n_tot)
        """
        return prob_n_hits_lt_n_des(self.n_hits_local, self.n_crits('local'))

    def n_hits_lt_n_des_local(self):
        """
        n_hits <= n_des
        """
        return self.n_hits_local <= self.n_crits('local')

    def heat_loads_local(self):
        return np.array([load_from_n_hits(
                            n_hits, self.n_tot, area, self.p_conv, self.ceta
                            )
                         if area != 0 else 0.
                         for n_hits, area in zip(self.n_hits_local,
                                                 self.areas_local)
                         ])


class TileBasedTracerEvaluator(TracerEvaluator):
    reshape = (-1, 12)

    def setup(self):
        super(TileBasedTracerEvaluator, self).setup()
        self._box_search_tree = None
        self._box_search_tree_path = None
        self._templates = None

    def clear(self):
        super(TileBasedTracerEvaluator, self).clear()
        self._mesh_hits = None

    @property
    def box_search_tree(self):
        if self._box_search_tree is None:
            if os.path.exists(self._box_search_tree_path):
                with open(self._box_search_tree_path, 'rb') as f:
                    if sys.version_info[0] == 2:
                        self._box_search_tree = pickle.load(f)
                    else:
                        # created with python2 thus latin1
                        self._box_search_tree = pickle.load(f,
                                                            encoding='latin1')
            else:
                tree = tfields.bounding_box.Searcher(self.mesh)
                if self._box_search_tree_path is not None:
                    with open(self._box_search_tree_path, 'wb') as f:
                        pickle.dump(tree, f)
                self._box_search_tree = tree

        return self._box_search_tree

    @property
    def mesh_hits(self):
        """
        tfields.Tensors that can be used as map field for self.mesh
        """
        if self._mesh_hits is None:
            mesh_hits = tfields.Tensors(
                self.box_search_tree.in_faces(self.hits, delta=None))
            mesh_hits = tfields.Tensors(mesh_hits.sum(axis=0))
            self.mesh_hits = mesh_hits  # use setter
        return self._mesh_hits

    @mesh_hits.setter
    def mesh_hits(self, mesh_hits):
        self._mesh_hits = mesh_hits

    @property
    def templates(self):
        if self._templates is None:
            raise ValueError("templates not set")
        return self._templates

    def tiles_with_maps_fields(self, merged=False):
        """
        Returns:
            tiled meshes with obj.maps[0].fields = [
                number of hits summed over a tile,
                number of hits summed over a tile per tile area,
                ]
        """
        templates = [t.copy() for t in self._tiles()]
        for t, n, a in zip(templates, self.n_hits, self.areas):
            t.maps[0].fields = [
                tfields.Tensors(np.full(len(t.maps[0]),
                                float(n))),
                tfields.Tensors(np.full(len(t.maps[0]),
                                float(n) / a if a != 0 else 0.)),
                ]
        if merged:
            return tfields.Mesh3D.merged(*templates)
        return templates

    def mesh_with_maps_fields(self):
        """
        Returns:
            meshe with obj.maps[0].fields = [
                number of hits per face,
                number of hits per face per face area,
                ]
        """
        mesh = self.mesh.copy()
        mesh_hits = self.mesh_hits / self.mesh.triangles().areas()
        mesh.maps[0].fields = [
            self.mesh_hits,
            mesh_hits,
            ]
        return mesh

    def _tiles(self):
        """
        Map the mesh to a tile geometry with the template
        provided for this
        """
        # convert to hits per area
        mesh_hits = self.mesh_hits / self.mesh.triangles().areas()
        self.mesh.maps[0].fields.append(mesh_hits)

        tiles = []
        for template in self.templates:
            tile = self.mesh.cut(template)
            if len(tile.maps) > 0 and len(tile.maps[0].fields) > 0:
                # convert hits per area back to hits. You will get fractions of
                # hits now.
                areas = tile.triangles().areas()
                areas[np.isnan(areas)] = 0.
                tile.maps[0].fields[0] *= areas
            tiles.append(tile)
        self.mesh.maps[0].fields = []
        return tiles

    def _eval_n_hits(self):
        tiles = self._tiles()
        n_hits = []
        for tile in tiles:
            if len(tile.maps) == 0:
                n = 0.
            else:
                n = float(tile.maps[0].fields[0].sum())
            n_hits.append(n)
        return n_hits

    def measurements_projected_tile_meshs(self):
        # NOTE: it could be possible to fill nan values with
        #       closest measurement value

        measurements_projected = self.measurements_projected

        measurement_tiles = []
        for template in self.templates:
            tile = measurements_projected.cut(template)
            if len(tile.maps) > 0 and len(tile.maps[0].fields) > 0:
                fields = []
                for i, field in enumerate(tile.maps[0].fields):
                    mask = ~np.isnan(field)
                    if any(mask):
                        field = np.full(len(field), np.mean(field[mask]))
                    else:
                        field = np.full(len(field), np.nan)
                    fields.append(field)
                tile.maps[0].fields = fields
                tile.maps[0].names = measurements_projected.maps[0].names
            measurement_tiles.append(tile)
        return measurement_tiles

    def measurements_projected_tile_values(self):
        """
        Returns:
            np.ndarray of shape n_mesurements, n_tiles:
                average measurments in each tile
        """
        tile_values = []
        measurements_projected_tile_meshs = self.measurements_projected_tile_meshs()
        n_fields = None
        for tile_mesh in measurements_projected_tile_meshs:
            if len(tile_mesh.maps) > 0 and len(tile_mesh.maps[0].fields) > 0:
                n_fields = len(tile_mesh.maps[0].fields)
                break
        for tile in measurements_projected_tile_meshs:
            if len(tile.maps) == 0:
                vs = np.full(n_fields, 0.)
            elif len(tile.maps[0].fields[0]) == 0:
                vs = np.full(n_fields, 0.)
            else:
                vs = np.array([tile.maps[0].fields[i][0]
                               for i in range(n_fields)])
            tile_values.append(vs)
        values = np.array(tile_values).T
        tf = tfields.TensorFields(
            np.empty(len(measurements_projected_tile_meshs)),
            *values)
        tf.names = self.measurements_projected.maps[0].names
        return tf


class SearchingSphereTracerEvaluator(TracerEvaluator):
    """
    Examples:
        Saving hits in blender manner:
            con_l = ...
            evaluator = w7x.evaluation.SearchingSphereTracerEvaluator()
            evaluator.connection_length = con_l
            mm_ids = getattr(w7x.MeshedModelsIds, 'baffle')
            evaluator.mm_ids = mm_ids

            radius = 0.056
            evaluator.radius = radius
            print(evaluator.n_hits)  # this line is necessary, so that n_hits is
            # calculated

            sphere = tfields.Mesh3D.grid((radius, radius, 1),
                                         (-np.pi, np.pi, 12),
                                         (-np.pi / 2, np.pi / 2, 12),
                                         coord_sys='spherical')
            sphere.transform('cartesian')
            for i in range(3):
                sphere[:, i] += evaluator.center[i]
            sphere.save('~/tmp/center.obj')

            oktaeder_radius = 0.001
            oktaeder = tfields.Mesh3D.grid((oktaeder_radius, oktaeder_radius, 1),
                                           (-np.pi, np.pi, 5),
                                           (-np.pi / 2, np.pi / 2, 3),
                                           coord_sys='spherical')
            oktaeder.transform('cartesian')
            point_meshes = []
            for point in evaluator.hits:
                big_point = oktaeder.copy()
                for i in range(3):
                    big_point[:, i] += point[i]
                point_meshes.append(big_point)
            point_meshes = tfields.Mesh3D.merged(*point_meshes)
            point_meshes.save('~/tmp/hit_points.obj')
    """
    def setup(self):
        super(SearchingSphereTracerEvaluator, self).setup()
        self._radius = None

    def clear(self):
        super(SearchingSphereTracerEvaluator, self).clear()
        self.center = None

    @property
    def radius(self):
        if self._radius is None:
            raise ValueError("radius not set")
        return self._radius

    @radius.setter
    def radius(self, radius):
        self._radius = radius

    @property
    def areas(self):
        """
        Segment areas
        """
        return [np.pi * self.radius**2]

    @areas.setter
    def areas(self, areas):
        raise ValueError("Not allowed to set the radius directly")

    def _eval_n_hits(self):
        if len(self.hits) == 0:
            self.center = [np.nan] * 3
            return [0]
        epsilon_neighbourhoods = self.hits.epsilon_neighbourhood(self.radius)
        n_in_sphere = [v.shape[0] for v in epsilon_neighbourhoods]
        max_index = n_in_sphere.index(max(n_in_sphere))
        self.center = self.hits[max_index]
        n_hits = n_in_sphere[max_index]
        return [n_hits]

    def plot(self, value, **kwargs):
        axis = kwargs.pop('axis', rna.plotting.gca(2))
        artist = axis.text(0.5, 0.5, value)
        return artist

    @property
    def areas_local(self):
        return self.areas

    @property
    def n_hits_local(self):
        self._n_hits_local = self.n_hits
        return super(SearchingSphereTracerEvaluator, self).n_hits_local



"""
Transcoding input.txt file.

Yet reading only
"""
import transcoding as tc
import logging


#  TODO(@dboe): Reading only
#  TODO(@amerlo): Make main() vmec_id works
#  TODO(@dboe): Make read function independent of blocks position in input list
def get_transcoding():
    def f_read_number_string(number_string, name, dtype=float):
        """
        Read lines like "abc = 1 1 2 3 5 8"
        """
        try:
            return {name: float(number_string)}
        except:
            return None

    def f_read_equal(line, name, event=None, dtype=float):
        """
        Read lines like "abc = 42.0"
        """
        if event is not None:
            if not event(line):
                return None
        _, _, number_string = line.rpartition('=')
        return f_read_number_string(number_string, name, dtype=dtype)

    #  Number of Poloidal Grid Points
    b_mpol = tc.Block(tc.Pattern("MPOL = {mpol:d}",
                                 read_method=tc.read_method(
                                     f_read_equal, 'mpol',
                                     lambda x: 'MPOL' in x)),
                      start=tc.Trigger(lambda x: 'MPOL' in x))

    #  NTHETA
    b_ntheta = tc.Block(tc.Pattern("NTHETA = {ntheta:d}",
                                   read_method=tc.read_method(
                                       f_read_equal, 'ntheta',
                                       lambda x: 'NTHETA' in x)),
                        start=tc.Trigger(lambda x: 'NTHETA' in x))

    #  Tolerance
    b_tolerance = tc.List(
        tc.Pattern("{forceToleranceLevels:f}",
                   read_method=tc.read_method(f_read_number_string,
                                              'forceToleranceLevels')),
        separator=' ',
        prefix=tc.Margin.alternatives(' FTOL_ARRAY = ', 'FTOL_ARRAY = '),
        rigid=False,
        start=tc.Trigger(lambda x: 'FTOL_ARRAY' in x))
    #  Phi edge
    b_phiEdge = tc.Block(tc.Pattern(
        "PHIEDGE = {maxToroidalMagneticFlux:f}",
        read_method=tc.read_method(f_read_equal, 'maxToroidalMagneticFlux',
                                   lambda x: 'PHIEDGE' in x)),
                         start=tc.Trigger(lambda x: 'PHIEDGE' in x))

    #  Pressure Scale
    b_pressureScale = tc.Block(tc.Pattern("PRES_SCALE = {pressureScale:f}",
                                   read_method=tc.read_method(
                                       f_read_equal, 'pressureScale',
                                       lambda x: 'PRES_SCALE' in x)),
                               start=tc.Trigger(lambda x: 'PRES_SCALE' in x),
                               barrier=tc.Trigger(lambda x: 'CURTOR' in x),
                               default={'pressureScale': 1.0})

    #  Total Toroidal Current
    b_curtor = tc.Block(tc.Pattern("CURTOR = {totalToroidalCurrent:f}",
                                   read_method=tc.read_method(
                                       f_read_equal, 'totalToroidalCurrent',
                                       lambda x: 'CURTOR' in x)),
                        start=tc.Trigger(lambda x: 'CURTOR' in x))

    #  Radial resolution
    t_pressureProfile = tc.Trigger(
        lambda x: "PMASS_TYPE" in x or "IPMASS" in x, delay=0)
    b_radial = tc.List(
        tc.Pattern("{numGridPointsRadial:d}",
                   read_method=tc.read_method(f_read_number_string,
                                              'numGridPointsRadial')),
        separator=' ',
        prefix=tc.Margin.alternatives(' NS_ARRAY = ', 'NS_ARRAY = ',
                                      'NS_ARRAY =   '),
        start=tc.Trigger(lambda x: 'NS_ARRAY' in x),
        rigid=False,
        barrier=t_pressureProfile)
    #  Pressure Profile
    t_pressureProfileBarrier = tc.Trigger(
        lambda x: "AM = " in x or " AM = " in x or " AM_AUX_F = " in x,
        delay=0)

    def f_pressureProfileHead(line):
        if 'PMASS_TYPE' in line:
            return {
                'pressureProfileProfileType':
                line.split('= ')[1].replace("'", "")
            }
        if "IPMASS" in line:
            lut = {
                0: 'power_series',
                1: 'two_Lorentz',
                11: 'Akima_spline',
                13: 'cubic_spline'
            }
            ip = int(line.split('=')[1])
            return {'pressureProfileProfileType': lut[ip]}
        return None

    b_pressureProfileHead = tc.Block(tc.Pattern(
        "PMASS_TYPE = '{pressureProfileProfileType}'",
        read_method=tc.read_method(f_pressureProfileHead)),
                                     start=t_pressureProfile,
                                     barrier=t_pressureProfileBarrier)
    b_pressureProfile = tc.List(
        tc.Pattern("{pressureProfileCoefficients:f}",
                   read_method=tc.read_method(f_read_number_string,
                                              'pressureProfileCoefficients')),
        separator=' ',
        rigid=False,
        prefix=tc.Margin.alternatives(' AM = ', 'AM = ', 'AM_AUX_F = '))

    #  Coil Currents
    b_coilCurrents = tc.List(
        tc.Pattern("{coilCurrents:f}",
                   read_method=tc.read_method(f_read_number_string,
                                              'coilCurrents')),
        separator=' ',
        prefix=tc.Margin.alternatives(' EXTCUR = ', 'EXTCUR = '),
        start=tc.Trigger(lambda x: 'EXTCUR' in x),
        default={'coilCurrents': []})

    #  Number of Iterations
    b_niter = tc.Block(tc.Pattern("NITER = {niter:d}",
                                  read_method=tc.read_method(
                                      f_read_equal, 'niter',
                                      lambda x: 'NITER' in x)),
                       start=tc.Trigger(lambda x: 'NITER' in x))

    #  Number of Steps
    b_nstep = tc.Block(tc.Pattern("NSTEP = {nstep:d}",
                                  read_method=tc.read_method(
                                      f_read_equal, 'nstep',
                                      lambda x: 'NSTEP' in x)),
                       start=tc.Trigger(lambda x: 'NSTEP' in x))

    #  Current Profile
    t_rAxis_start = tc.Trigger(lambda x: "RAXIS" in x, delay=0)

    def f_currentProfileHead(line):
        if 'PCURR_TYPE' in line:
            return {
                'currentProfileProfileType':
                line.split('= ')[1].replace("'", "")
            }
        if "IPCURR" in line:
            lut = {
                0: 'power_series',
                1: 'sum_atan',
                2: 'sum_atan',
                11: 'Akima_spline_l',
                12: 'Akima_spline_lp',
                13: 'cubic_spline_l',
                14: 'cubic_spline_lp'
            }
            ip = int(line.split('=')[1])
            return {'currentProfileProfileType': lut[ip]}
        return None

    b_currentProfileHead = tc.Block(
        tc.Pattern("PCURR_TYPE = '{currentProfileProfileType}'",
                   read_method=tc.read_method(f_currentProfileHead)),
        start=tc.Trigger(lambda x: "PCURR_TYPE" in x or "IPCURR" in x,
                         delay=0),
        barrier=tc.Trigger(
            lambda x: "AC = " in x or " AC = " in x or " AC_AUX_F = " in x,
            delay=0),
        default={'currentProfileProfileType': 'power_series'})
    b_currentProfile = tc.List(
        tc.Pattern("{currentProfileCoefficients:f}",
                   read_method=tc.read_method(f_read_number_string,
                                              'currentProfileCoefficients')),
        separator=' ',
        prefix=tc.Margin.alternatives('AC = ', ' AC = ', 'AC_AUX_F = '),
        rigid=False,
        barrier=t_rAxis_start,
        default={'currentProfileCoefficients': []})

    #  rAxis
    b_rAxis = tc.List("{RAXIS:f}",
                      separator=' ',
                      prefix=tc.Margin.alternatives("{tmp:5} = ",
                                                    " {tmp:5} = "),
                      rigid=False,
                      start=t_rAxis_start)

    #  zAxis
    b_zAxis = tc.List("{ZAXIS:f}",
                      separator=' ',
                      prefix=tc.Margin.alternatives("{tmp:5} = ",
                                                    " {tmp:5} = "),
                      rigid=False,
                      start=tc.Trigger(lambda x: "ZAXIS" in x, delay=0))

    #  Boundary
    def f_boundary(line):
        if 'rbc' in line or 'RBC' in line:
            splt = line.split(',')
            if '' in splt:
                splt.remove('')
            if len(splt) != 4:
                return None
            vals = {
                'n': int(splt[0].split('(')[1]),
                'm': int(splt[1].split(')')[0]),
                'rbc': float(splt[1].split('=')[1]),
                'nz': int(splt[2].split('(')[1]),
                'mz': int(splt[3].split(')')[0]),
                'zbs': float(splt[3].split('=')[1])
            }
            return vals
        return None

    b_boundary = tc.Table(tc.Pattern(
        " RBC({n:d},{m:d}) = {rbc:g}  ZBS({nz:d},{mz:d}) = {zbs:g}",
        read_method=tc.read_method(f_boundary)),
                          stop=tc.Trigger(lambda x: "/" in x, delay=0))

    class CorrectedTranscoding(tc.Transcoding):
        def read(self, inp):
            newInp = []
            for line in inp:
                if "ZAXIS" in line:
                    if 'rbc' in line:
                        newInp.append(line.split('rbc')[0])
                        newInp.append('rbc' + line.split('rbc')[1])
                    else:
                        break
                newInp.append(line)
            else:
                inp = newInp
            d = super(CorrectedTranscoding, self).read(inp)
            if "pressureProfileProfileType" not in d:
                d["pressureProfileProfileType"] = 'power_series'
            if "currentProfileProfileType" not in d:
                d["currentProfileProfileType"] = 'power_series'
            return d

    transcoding = CorrectedTranscoding(
        b_coilCurrents, b_niter, b_nstep, b_tolerance, b_phiEdge, b_mpol,
        b_ntheta, b_radial, b_pressureProfileHead, b_pressureProfile,
        b_currentProfileHead, b_currentProfile, b_pressureScale, b_curtor, b_rAxis, b_zAxis,
        b_boundary)
    return transcoding


if __name__ == '__main__':
    import w7x
    import logging
    log = logging.getLogger()
    log.setLevel(1)
    run = w7x.vmec.Run("w7x_ref_9")
    run._parse_input()
    print(run._input)



import unittest
import w7x
import random
import numpy as np


class TestDistribution(unittest.TestCase):
    def test_sample_and_mean(self):
        d = w7x.batch.Distribution(random.randint, 0, 10)
        samples = [next(d) for i in range(100)]
        self.assertTrue(np.mean(samples) > 0 and np.mean(samples) < 10)

    def test_fixed_distribution(self):
        d = w7x.batch.Distribution(random.randint, 10, 10)
        samples = [next(d) for i in range(100)]
        self.assertTrue(samples == [10] * len(samples))


if __name__ == '__main__':
    unittest.main()



"""
specific methods for poincare plotting
"""
import tfields


def plot_poincare_surfaces(poincareSurfaces, **kwargs):
    """
    Args:
        poincareSurfaces (list of Points3D): each Points3D instance is one
            fieldLine followed around the torus
    """
    methodName = kwargs.pop('methodName', 'scatter')
    po = tfields.plotting.PlotOptions(kwargs)
    rMin = po.pop('rMin', 4.0)
    rMax = po.pop('rMax', 6.6)
    zMin = po.pop('zMin', -1.3)
    zMax = po.pop('zMax', +1.3)
    phiRad = po.pop('phiRad', None)

    po.set_default('yAxis', 2)
    po.set_default('labels', ['r (m)', r'$\phi$ (rad)', 'z (m)'])
    if methodName is 'scatter':
        po.set_default('marker', '.')
        po.set_default('s', 1)
    po.set_default('methodName', methodName)
    color_given = True
    if 'color' not in po.plot_kwargs:
        color_given = False
        cmap, _, _ = po.getNormArgs()
        color_cycle = tfields.plotting.color_cycle(cmap, len(poincareSurfaces))
    elif isinstance(po.get('color'), list):
        color_given = False  # hack to set the color from list
        color_cycle = iter(po.get('color'))
    artists = []
    for i, surfacePoints in enumerate(poincareSurfaces):
        with surfacePoints.tmp_transform(tfields.bases.CYLINDER):
            phiSurface = surfacePoints[:, 1]
            if phiRad is None:
                phiRad = phiSurface[0]
            if bool((phiSurface != phiRad).any()):
                continue
            if not color_given:
                po.set('color', next(color_cycle))
            artists.append(surfacePoints.plot(axis=po.axis, **po.plot_kwargs))
    tfields.plotting.set_aspect_equal(po.axis)
    po.axis.set_xlim(rMin, rMax)
    po.axis.set_ylim(zMin, zMax)
    return artists


def plot_poincare_geometries(geometries, **kwargs):
    po = tfields.plotting.PlotOptions(kwargs)
    po.set_default('methodName', 'plot')
    po.set_default('lw', 1)
    artists = []
    for p in range(len(geometries)):
        for g in range(len(geometries[p])):
            artists.extend(plot_poincare_surfaces(geometries[p][g], axis=po.axis,
                                                **po.plot_kwargs)) 
    return artists
